/***************************************************************************
                          salsa.cpp  -  description
                             -------------------
    begin                : Mon Mr 10 2003
    copyright            : (C) 2003 by Ingmar Bauer
    email                : ingmar@metacrawler.de
 ***************************************************************************/

using namespace std;

#include "matrix.h"
#include "datasource.h"
#include "matrixsource.h"
#include "salsa.h"

//#define DEBUG
//#define SALSA_DAMPENING

/****************************************************************************/
/*** salsa::salsa(datasource &ds)                                         ***/
/***----------------------------------------------------------------------***/
/*** Constructor to create Hub- und Authority matrix. Needs Index-        ***/
/*** datasource and saves matrix-files.                                   ***/
/****************************************************************************/
salsa::salsa(datasource &ds){
  this->ds = &ds;
  this->size = this->ds->getMatrixSize();
}

/****************************************************************************/
/*** salsa::salsa(matrixsource &ms)                                       ***/
/***----------------------------------------------------------------------***/
/*** Constructor to use existing Hub- und Authority matrices for          ***/
/*** calculation of ranking.                                              ***/
/****************************************************************************/
salsa::salsa(matrixsource &ms, float c=0.85, float eps=0.01) {
  this->ms = &ms;
  this->c = c;
  this->eps = eps;
}

/****************************************************************************/
/*** calc_eigenvector()                                                   ***/
/****************************************************************************/
vector<float> salsa::calc_eigenvector() {
  vector<float> src;
  vector<float> dest;
  int n_el;
  int i;
  matrix_element el;
  float s=eps+1;  // vector distance
  int line=0;

  // init vectors
  src.clear();
  for(i=0; i<ms->getSize(); i++) {
    src.insert(src.begin(), 1, i+1);
  }

  // normalize vector
  float sum=0.0;
  for(i=0;i<ms->getSize(); i++) {
    sum += (src[i]*src[i]);
  }
  for(i=0;i<ms->getSize(); i++) {
    src[i] /= sqrt(sum);
  }

  // start computation
  while(s > eps) {

    dest.clear();
    dest.insert(dest.begin(), ms->getSize(), 0.0);

    // let's start computing dominant eigenvector
    ms->reInitMatrix();

    int stop=0;

    do {
      n_el = ms->prepareNextRow(line);
      if(n_el == 0) {
        stop=1;
      }

      while(ms->getNextRowEl(el) > 0) {
        dest[line] += (src[el.col]*el.value);
      }
    } while(!stop);

    // dampening
    #ifdef SALSA_DAMPENING
    #ifdef DEBUG
    cout << "dampening... (c=" << c << ")\n";
    #endif
    #endif

    // normalize vector
    float sum=0.0;
    for(i=0;i<ms->getSize(); i++) {
      sum += (dest[i]*dest[i]);
    }
    for(i=0;i<ms->getSize(); i++) {
      dest[i] /= sqrt(sum);
    }

    // calculate vector distance
    #ifdef DEBUG
    cout << "calculating vector distance...\n";
    #endif

    s = 0.0;            // vector distance
    for(i=0; i<ms->getSize(); i++) {
      s+=fabs(dest[i]-src[i]);
    }

    cout << "Vectorabstand: " << s << endl;

    // loop
    src = dest;
  }

  return dest;
}

vector<float> salsa::calc() {
  vector<float> src;
  vector<float> dest;
  int n_el;
  int i;
  matrix_element el;
  float s=eps+1;  // vector distance
  int line=0;

  // init vectors
  src.clear();
  for(i=0; i<ms->getSize(); i++) {
    src.insert(src.begin(), 1, i+1);
  }

  while(s > eps) {

    dest.clear();
    dest.insert(dest.begin(), ms->getSize(), 0.0);

    // let's start computing dominant eigenvector
    ms->reInitMatrix();

    int stop=0;
    
    do {
      n_el = ms->prepareNextRow(line);
      if(n_el == 0) {
        stop=1;
      }

      while(ms->getNextRowEl(el) > 0) {
        dest[line] += (src[el.col]*el.value);
      }
    } while(!stop);

    // dampening
    #ifdef DEBUG
    cout << "dampening... (c=" << c << ")\n";
    #endif
    
    for(i=0;i<ms->getSize();i++) {
      dest[i] = c * dest[i] + ((1.0-c)/ms->getSize()); // dampening
    }

    // calculate vector distance
    #ifdef DEBUG
    cout << "calculating vector distance...\n";
    #endif
    
    s = 0.0;            // vector distance
    for(i=0; i<ms->getSize(); i++) {
      s+=fabs(dest[i]-src[i]);
    }

    cout << "Vectorabstand: " << s << endl;

    // loop
    src = dest;
  }

  return dest;
}

salsa::~salsa(){
#ifdef DEBUG
  cout << "salsa::~salsa() - destruction time" << endl;
#endif
}

/****************************************************************************/
/*** int salsa::createMatrices()                                          ***/
/***----------------------------------------------------------------------***/
/*** Erzeugt Hub- und Authority Matrix anhand eines vorhandenen Index mit ***/
/*** Hilfe der Klasse datasource.                                         ***/
/****************************************************************************/
int salsa::createMatrices() {

  adj = matrix<float>(size);
  ds->getAdjascencyMatrix(adj);

  // create hub and authority matrix
  hub = matrix<float>(size);
  auth = matrix<float>(size);

  int x,y;

  // lookup-vector for column sum
  vector<float> sum_row(size);
  vector<float> sum_col(size);

  #ifdef DEBUG
  cout << "calculating row/col sums" << endl;
  #endif
  
  for(y=0;y<size;y++) { // for every column
    sum_row[y] = 0.0;
    sum_col[y] = 0.0;
    for(x=0;x<size;x++) { // summarize all values
      sum_col[y]+=adj.m[x][y];
      sum_row[y]+=adj.m[y][x];
      //cout << adj.m[y][x] << ", ";

      hub.m[y][x] = auth.m[y][x] = 0.0;
    }
  }

  #ifdef DEBUG
  cout << "Starting computation of hab and authority matrix (size: " << size << ")" << endl;
  #endif
  
  // start operation
  for(y=0;y<size;y++) {
    for(x=0;x<size;x++) {
      // multiplication
      for(int i=0; i<size; i++) {
        if(sum_row[y]>0 && sum_col[i]>0) {
          hub.m[y][x] += ( (adj.m[y][i]/sum_row[y]) * (adj.m[x][i]/sum_col[i]));
        }
        if(sum_col[y]>0 && sum_row[i]>0) {
          auth.m[y][x] += ( (adj.m[i][x]/sum_col[y]) * (adj.m[i][y]/sum_row[i]) );
        }
      }
    }
  }

  #ifdef DEBUG
  for(y=0;y<size;y++) {
    for(x=0;x<size;x++) {
      cout << hub.m[y][x] << ", ";
    }
    cout << endl;
  }

  cout << endl << endl;

  for(y=0;y<size;y++) {
    for(x=0;x<size;x++) {
      cout << auth.m[y][x] << ", ";
    }
    cout << endl;
  }
  #endif
  
  return 1;
}

int salsa::saveMatrices(string fname) {
  hub.save(fname+string(".hub"));
  auth.save(fname+string(".aut"));
  return 1;
}
